#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: analytic_geometry_base .py
# @time: 2020/6/5 4:24
# @Software: PyCharm


from abc import ABC

__all__ = [
    'Line',
    'Surface'
]


class Line(ABC):
    pass


class Surface(ABC):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.snowland.ltd
# @file: color.py
# @time: 2018/7/26 0:24
# @Software: PyCharm

import numpy as np

npa = np.array


def rgb2ycbcr(img):
    origT = npa([[65.481, 128.553, 24.966], [-37.797, -74.203, 112], [112, -93.786, -18.214]])
    oriOffset = npa([[16], [128], [128]])

    # if img.dtype.name == 'uint8':
    if isinstance(img, np.uint8):
        t = 1
        offset = 1.0 / 255
    # elif img.dtype.name == 'float64':
    elif isinstance(img, (np.float, np.float64, float)):
        t = 1.0 / 255
        offset = 1.0 / 255
    # elif img.dtype.name == 'uint16':
    elif isinstance(img, np.uint16):
        t = 257.0 / 65535
        offset = 257
    else:
        raise ValueError('image type not support ')
    T = origT * t
    Offset = oriOffset * offset

    ycbcr = np.zeros(img.shape, dtype=img.dtype)
    for p in range(3):
        ycbcr[:, :, p] = T[p, 0] * img[:, :, 0] + T[p, 1] * img[:, :, 1] + T[p, 2] * img[:, :, 2] + Offset[p]

    return ycbcr


def ycbcr2rgb(img):
    origT = npa([[65.481, 128.553, 24.966], [-37.797, -74.203, 112], [112, -93.786, -18.214]])
    oriOffset = npa([[16], [128], [128]])
    tinv = np.linalg.inv(origT)
    # if img.dtype.name == 'uint8':
    if isinstance(img, np.uint8):
        t = 255
        offset = 255
    elif isinstance(img, (np.float, np.float64)):
    # elif img.dtype.name == 'float64':
        t = 255
        offset = 1
    # elif img.dtype.name == 'uint16':
    elif isinstance(img, np.uint16):
        t = 65535 / 257.0
        offset = 65535
    else:
        raise ValueError('image type not support ')
    T = tinv * t
    Offset = offset * tinv.dot(oriOffset)

    rgb = np.zeros(img.shape, dtype=img.dtype)

    for p in range(3):
        rgb[:, :, p] = T[p, 0] * img[:, :, 0] + T[p, 1] * img[:, :, 1] + T[p, 2] * img[:, :, 2] - Offset[p]

    if isinstance(img, (float, np.float64)):
        rgb[rgb > 1.0] = 1.0
        rgb[rgb < 0] = 0
    return rgb



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: qgis_utils.py
# @time: 2022/01/06 20:52
# @Software: PyCharm


__auth__ = 'A.Star'

from typing import Tuple, List

import numpy as np

npa = np.array


EARTH_REDIUS = 6378.137  # 初始化地球半径km
try:
    from qgis.core import QgsDistanceArea, QgsUnitTypes, QgsPointXY
    Ellipsoid = (6378137, 6356752.314245179)
    disClass = QgsDistanceArea()
    disClass.setEllipsoid(Ellipsoid[0], Ellipsoid[1])
    qgis_import = True
except:
    qgis_import = False


def haversine(Olon, Olat, Dlon, Dlat):
    """
    通过两经纬度点计算地表实际距离
    """
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = EARTH_REDIUS * c
    return d * 100000  # 转换到cm为单位


def haversine_metres(Olon, Olat, Dlon, Dlat):
    return haversine(Olon, Olat, Dlon, Dlat) / 100


def length(Olon, Olat):
    """
    通过两经纬度点计算地表实际距离

    """
    return np.sum(haversine(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:]))


def length_metres(Olon, Olat):
    """
    通过两经纬度点计算地表实际距离
    """
    return np.sum(haversine(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:])) / 100


def haversine_by_points(p1, p2):
    """
    通过两经纬度点计算地表实际距离

    """
    Olon, Olat = p1[:, 0], p1[:, 1]
    Dlon, Dlat = p2[:, 0], p2[:, 1]
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = EARTH_REDIUS * c
    return d * 100000  # 转换到cm为单位


def haversine_metres_by_points(p1, p2):
    return haversine_by_points(p1, p2) / 100


def haversine_metres_by_two_points(p1, p2):
    if qgis_import:
        p1, p2 = QgsPointXY(*p1), QgsPointXY(*p2)
        return disClass.convertLengthMeasurement(disClass.measureLine(p1, p2), QgsUnitTypes.DistanceMeters)
    else:
        Olon, Olat = p1
        Dlon, Dlat = p2
        d_lat = np.radians(Dlat - Olat)
        d_lon = np.radians(Dlon - Olon)
        a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
             np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
             np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
        c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
        d = EARTH_REDIUS * c
        return d * 1000


def haversine_by_two_points(p1:(np.ndarray, Tuple, List), p2:(np.ndarray, Tuple, List)):
    """
    通过两经纬度点计算地表实际距离
    """
    Olon, Olat = p1
    Dlon, Dlat = p2
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = EARTH_REDIUS * c
    return d * 100000  # 转换到cm为单位





import numpy as np


class GisHelper:

    @staticmethod
    def get_point_by_rate(line: np.ndarray, meters):
        return GisHelper.get_point_by_rate_index(line, meters)[0]

    @staticmethod
    def get_point_by_rate_index(line: np.ndarray, meters):
        meters_all = [meters(a, b) for a, b in zip(line[:-1], line[1:])]
        s = np.cumsum(meters_all)
        if s[-1] < meters:
            return line[-1], len(s)
        ind = np.sum(s < meters)
        s = np.insert(s, 0, 0)
        delta_meters = meters - s[ind]
        vector_line = line[1:] - line[:-1]
        return line[ind] + delta_meters / meters_all[ind] * vector_line[ind], ind



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: 
# @file: __init__.py.py
# @time: 2018/7/26 10:31
# @Software: PyCharm


from snowland.image.color import *
from snowland.image.api import *
from snowland.image.transform import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: 
# @file: img_to_sketch.py
# @time: 2018/8/6 1:15
# @Software: PyCharm

from skimage.color import rgb2grey
import numpy as np


def sketch(img, threshold=15):
    """
    素描画生成
    param img: Image实例
  param threshold: 介于0到100
    :return:
    """
    if threshold < 0:
        threshold = 0
    if threshold > 100:
        threshold = 100
    if len(img.shape) == 3:
        img = rgb2grey(img)
    m, n = img.shape
    diff = np.abs(img[:m - 1, :n - 1] - img[1:, 1:])
    img = np.zeros((m - 1, n - 1))
    img[diff < threshold/255] = 1
    return img



# -*- coding: utf-8 -*-

import numpy as np

npl = np.linalg
npa = np.array


def DMS2NDS(dms):
    nds = int((1 << 32) / 360 * dms)
    return nds


def NDS2Morton(ndsLon, ndsLat):
    bit = 1
    mortonCode = 0
    x, y = ndsLon, ndsLat

    if y < 0:
        y += 0x7FFFFFFF

    y <<= 1

    for i in range(32):
        mortonCode |= x & bit
        x <<= 1
        bit <<= 1
        mortonCode |= y & bit
        y <<= 1
        bit <<= 1

    return mortonCode


def get_tile_ID(lon, lat, level):
    ndsLon = DMS2NDS(lon)
    ndsLat = DMS2NDS(lat)
    morton = NDS2Morton(ndsLon, ndsLat)
    nTileID = (((morton >> (2 * (31 - level))) & 0xffffffff) + (1 << (16 + level)))
    return nTileID



# !/usr/bin/env python
#  -*- coding: utf-8 -*-
#  @Author  : 河北雪域网络科技有限公司 A.Star
#  @contact: astar@snowland.ltd
#  @site: 
#  @file: main.py
#  @time: 2018/7/11 14:34
#  @Software: PyCharm

import cv2


def cartoonise(img_rgb, num_down=2, num_bilateral=7):
    """
    :param img_rgb: 输入图像
    :param num_down: 缩减像素采样的数目
    :param num_bilateral: 定义双边滤波的数目
    :return:
    """
    # 用高斯金字塔降低取样
    img_color = img_rgb
    for _ in range(num_down):
        img_color = cv2.pyrDown(img_color)
    # 重复使用小的双边滤波代替一个大的滤波
    for _ in range(num_bilateral):
        img_color = cv2.bilateralFilter(img_color, d=9, sigmaColor=9, sigmaSpace=7)
    # 升采样图片到原始大小
    for _ in range(num_down):
        img_color = cv2.pyrUp(img_color)
    # 转换为灰度并且使其产生中等的模糊
    img_gray = cv2.cvtColor(img_color, cv2.COLOR_RGB2GRAY)
    img_blur = cv2.medianBlur(img_gray, 7)
    # 检测到边缘并且增强其效果
    img_edge = cv2.adaptiveThreshold(img_blur, 255,
                                     cv2.ADAPTIVE_THRESH_MEAN_C,
                                     cv2.THRESH_BINARY,
                                     blockSize=9,
                                     C=2)
    # 转换回彩色图像
    img_edge = cv2.cvtColor(img_edge, cv2.COLOR_GRAY2RGB)
    img_cartoon = cv2.bitwise_and(img_color, img_edge)
    return img_cartoon



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: utils.py
# @time: 2022/01/06 20:48
# @Software: PyCharm


import numpy as np
from scipy.spatial.distance import cdist, pdist, euclidean
from astartool.number import equals_zero

npa = np.array
npl = np.linalg


def middle(x, x1, x2, eps=1e-8):
    """
    判断点是否在x1, x2之间
    x, x1, x2 均为ndarray
    """
    dx1 = x - x1
    dx2 = x - x2
    return ((x1 < x) & (x < x2)) | ((x1 > x) & (x > x2)) | ((-eps < dx1) & (dx1 < eps)) | ((-eps < dx2) & (dx2 < eps))


def get_lines(line_points):
    """
    """
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]
    return np.vstack((A, B, C)).T * 10000


def get_foot(point, line=None, A=None, B=None, C=None):
    """
    point: 单点, ndarray/list len(points) = 2
    line: Ax+By+C=0
    返回 foot_x, foot_y, 垂足坐标

    """
    if line is not None:
        A, B, C = line[:, 0], line[:, 1], line[:, 2]
    elif A is not None and B is not None and C is not None:
        pass

    x, y = point[0], point[1]
    A2, B2, AB, AC, BC = A * A, B * B, A * B, A * C, B * C
    # foot_x, foot_y 即为垂足坐标
    foot_x = (B2 * x - AB * y - AC) / (A2 + B2)
    foot_y = (A2 * y - AB * x - BC) / (A2 + B2)

    return foot_x, foot_y


def get_intersect_point(line):
    """
    line_points 是 n x 3 矩阵，代表组成线标准式的参数 Ax+By+C=0 的系数 ABC
    返回这些线的交点
    """
    line1 = line[:-1, :]
    line2 = line[1:, :]

    a1 = line1[:, 0]
    b1 = line1[:, 1]
    c1 = line1[:, 2]
    a2 = line2[:, 0]
    b2 = line2[:, 1]
    c2 = line2[:, 2]
    b1c2 = b1 * c2
    b2c1 = b2 * c1
    a1c2 = a1 * c2
    a1b2 = a1 * b2
    a2b1 = a2 * b1
    a2c1 = a2 * c1

    x = (b1c2 - b2c1) / (a1b2 - a2b1)
    y = (a2c1 - a1c2) / (a1b2 - a2b1)
    return x, y


def distance_point_line(points, line_points, metric=euclidean):
    """
    points: 单点, ndarray/list len(points) = 2
    line_points: 每两点组成的一条线， 代表一条linestring
    返回tuple(isVe, ind, d)
    isVe: 找到的最近点是不是线的顶点
    p: 距离最近的点，ndarray
    d: 最近距离是多少
    """
    # ABC即为线的参数 Ax + By + C = 0
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]

    foot_x, foot_y = get_foot(points, A=A, B=B, C=C)
    # 是否为顶点坐标
    is_vertex = True

    index = middle(foot_x, line_points[1:, 0], line_points[:-1, 0]) & middle(foot_y, line_points[1:, 1],
                                                                             line_points[:-1, 1])
    if np.any(index):
        is_vertex = False
        points_check = np.vstack((foot_x[index].T, foot_y[index].T)).T
    else:
        points_check = line_points

    dist = cdist(npa([points]), points_check, metric=metric)
    ind = np.argmin(dist)
    return is_vertex, points_check[ind], dist[0, ind]


def distance_point_line_index(point, line_points, metric=euclidean):
    """
    points: 单点(x, y), ndarray/list len(points) = 2
    line_points: 每两点组成的一条线， 代表一条linestring
    返回tuple(isVe, ind, d)
    is_vertex: 找到的最近点是不是线的顶点
    p: 距离最近的点，ndarray
    d: 最近距离是多少
    ind: index
    """
    # ABC即为线的参数 Ax + By + C = 0
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]

    foot_x, foot_y = get_foot(point, A=A, B=B, C=C)
    # 是否为顶点坐标
    is_vertex = True

    index = middle(foot_x, line_points[1:, 0], line_points[:-1, 0]) & middle(foot_y, line_points[1:, 1],
                                                                             line_points[:-1, 1])
    if np.any(index):
        is_vertex = False
        index_ind, = np.where(index)
        points_check = np.vstack((foot_x[index].T, foot_y[index].T)).T

        dist = cdist(npa([point]), points_check, metric=metric)
        ind = np.argmin(dist)
        return is_vertex, points_check[ind], dist[0, ind], index_ind[ind]
    else:
        points_check = line_points
        dist = cdist(npa([point]), points_check, metric=metric)
        ind = np.argmin(dist)
        return is_vertex, points_check[ind], dist[0, ind], ind


def move_distance(a, dist, flag):
    """
    移线
    """
    lines = get_lines(a)
    norm_line = npl.norm(lines[:, :2], axis=1)
    delta_c = dist * norm_line
    lines_new = lines
    if flag:
        lines_new[:, 2] -= delta_c
    else:
        lines_new[:, 2] += delta_c
    x, y = get_intersect_point(lines_new)
    if len(x):
        x_start, y_start = get_foot(a[0, :], lines[:1, :])
        x_end, y_end = get_foot(a[-1, :], lines[-1:, :])
        x = np.hstack((x_start, x, x_end))
        y = np.hstack((y_start, y, y_end))
    return x, y


def move_distance_with_endpoints(a, dist, flag):
    """
    移线，注意和move_distance移线不同！！！
    此函数带有端点
    """
    lines = get_lines(a)
    norm_line = npl.norm(lines[:, :2], axis=1)
    delta_c = dist * norm_line
    lines_new = lines
    if flag:
        lines_new[:, 2] -= delta_c
    else:
        lines_new[:, 2] += delta_c
    x, y = get_intersect_point(lines_new)
    x_start, y_start = get_foot(a[0, :], lines[:1, :])
    x_end, y_end = get_foot(a[-1, :], lines[-1:, :])
    x = np.hstack((x_start, x, x_end))
    y = np.hstack((y_start, y, y_end))

    return x, y


def alignment(linestring1: np.ndarray, linestring2: np.ndarray, metric=euclidean, eps=1):
    """
    化简车道线,使之按垂直方向对齐
    linestring1: 第一根线 m x 2 ndarray
    linestring2: 第二根线 n x 2 ndarray
    metric: 距离计算函数
    eps: 若距离小于 eps, 则不需要截取
    """
    assert linestring1.shape[1] == 2
    assert linestring2.shape[1] == 2
    is_vertex, points_check, dist, ind = distance_point_line_index(linestring2[0], linestring1)
    if (not is_vertex) and metric(points_check, linestring1[0]) > eps:
        linestring1 = linestring1[ind + 1:]
        linestring1 = np.vstack((points_check, linestring1))
    else:
        is_vertex, points_check, dist, ind = distance_point_line_index(linestring1[0], linestring2, metric)
        if (not is_vertex) and metric(points_check, linestring2[0]) > eps:
            linestring2 = linestring2[ind + 1:]
            linestring2 = np.vstack((points_check, linestring2))

    is_vertex, points_check, dist, ind = distance_point_line_index(linestring2[-1], linestring1, metric)
    if (not is_vertex) and metric(points_check, linestring1[-1]) > eps:
        linestring1 = linestring1[:ind + 1]
        linestring1 = np.vstack((linestring1, points_check))
    else:
        is_vertex, points_check, dist, ind = distance_point_line_index(linestring1[-1], linestring2, metric)
        if (not is_vertex) and metric(points_check, linestring2[-1]) > eps:
            linestring2 = linestring2[:ind + 1]
            linestring2 = np.vstack((linestring2, points_check))

    return linestring1, linestring2


def rotate_geometry(poly: np.ndarray, rad):
    matrix = np.asarray([[np.cos(rad), np.sin(rad)],
                         [-np.sin(rad), np.cos(rad)]])
    new_poly = np.dot(poly, matrix)
    return new_poly


def bounding_box(poly: np.ndarray, eps=1e-10):
    """
    boundingBox
    """
    # shape of min_rect: [4, 2]
    x_min, y_min = np.min(poly, axis=0)
    x_max, y_max = np.max(poly, axis=0)
    min_rect = np.asarray([[x_min, y_min],
                           [x_max, y_min],
                           [x_max, y_max],
                           [x_min, y_max]])
    area = (x_max - x_min) * (y_max - y_min)
    return min_rect, area


def min_rotate_rect_a(hull: np.ndarray, eps=1e-10):
    """
    最小外接矩形--按面积
    """
    area = np.inf
    lines_com = []
    if equals_zero(hull[0] - hull[-1], eps):
        points_hull = np.vstack((hull, hull[0]))
    else:
        points_hull = hull

    lines = get_lines(points_hull)
    for i, p2 in hull:
        p1 = hull[i - 1, :]  # 凸包上两个点
        line = lines[i]
        dist_cmp = (points_hull[:, 0] * line[0] + points_hull[:, 1] * line[1] + line[2])
        dist = dist_cmp / npl.norm(line[:2])

        h_ind = np.argmax(np.abs(dist))  # 得到距离最大的点距离，即为高，同时得到该点坐标
        h = dist[h_ind]
        line_px = line
        line_px[2] += dist[h_ind]
        p = hull[h_ind]
        line_cz = npa([line[1], -line[0], -line[1] * p[0] + line[0] * p[1]])

        dist_cmp = (points_hull[:, 0] * line_cz[0] + points_hull[:, 1] * line_cz[1] + line[2])
        dist = dist_cmp / npl.norm(line_cz[:2])

        v_ind_max = np.argmax(dist)  # 得到距离最大的点距离，即为高
        v_ind_min = np.argmin(dist)

        w = np.abs(dist[v_ind_max]) + np.abs(dist[v_ind_min])
        a = h * w
        if area >= h * w:
            # 使面积最小
            area = a
            line_cz_1 = line_cz
            line_cz_1[2] += dist[v_ind_max]
            line_cz_2 = line_cz
            line_cz_2[2] += dist[v_ind_min]
            lines_com = [line, line_cz_1, line_px, line_cz_2, line]
    return get_intersect_point(lines_com)


def min_rotate_rect_c(hull: np.ndarray, eps=1e-10):
    """
    最小外接矩形--按周长
    """
    lines_com = []
    cir = np.inf
    if equals_zero(hull[0] - hull[-1], eps):
        points_hull = np.vstack((hull, hull[0]))
    else:
        points_hull = hull

    lines = get_lines(points_hull)
    for i, p2 in hull:
        p1 = hull[i - 1, :]  # 凸包上两个点
        line = lines[i]
        dist_cmp = (points_hull[:, 0] * line[0] + points_hull[:, 1] * line[1] + line[2])
        dist = dist_cmp / npl.norm(line[:2])

        h_ind = np.argmax(np.abs(dist))  # 得到距离最大的点距离，即为高，同时得到该点坐标
        h = dist[h_ind]
        line_px = line
        line_px[2] += dist[h_ind]
        p = hull[h_ind]
        line_cz = npa([line[1], -line[0], -line[1] * p[0] + line[0] * p[1]])

        dist_cmp = (points_hull[:, 0] * line_cz[0] + points_hull[:, 1] * line_cz[1] + line[2])
        dist = dist_cmp / npl.norm(line_cz[:2])

        v_ind_max = np.argmax(dist)  # 得到距离最大的点距离，即为高
        v_ind_min = np.argmin(dist)

        w = np.abs(dist[v_ind_max]) + np.abs(dist[v_ind_min])
        c = h + w
        if cir > h + w:
            # 使面积最小
            cir = c
            line_cz_1 = line_cz
            line_cz_1[2] += dist[v_ind_max]
            line_cz_2 = line_cz
            line_cz_2[2] += dist[v_ind_min]
            lines_com = [line, line_cz_1, line_px, line_cz_2, line]
    return get_intersect_point(lines_com)


def min_rotate_rect(hull: np.ndarray, cmp: str = 'a', eps=1e-10):
    """
    最小外接矩形
    """
    if cmp.lower().startswith('a'):
        return min_rotate_rect_a(hull, eps)
    else:
        return min_rotate_rect_c(hull, eps)




from qgis.core import QgsGeometry
from snowland.qgis_tool.qgis_util import vector, vector_end, vector_start, points, pointsXY, start_pointXY, end_pointXY

QgsGeometry.points = points
QgsGeometry.pointsXY = pointsXY
QgsGeometry.vector = vector
QgsGeometry.vector_start = vector_start
QgsGeometry.vector_end = vector_end
QgsGeometry.start_pointXY = start_pointXY
QgsGeometry.end_pointXY = end_pointXY




# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: 
# @time: 
# @Software: PyCharm


from pencildraw import pencil_drawing
from pencildraw.pencils import pencil0, pencil1,pencil2, pencil3, pencil4
from snowland.image.style_migration.img_to_cartoon import cartoonise
from snowland.image.style_migration.img_to_freehand import freehand
from snowland.image.style_migration.img_to_oilpaint import oilpaint
from snowland.image.style_migration.img_to_sketch import sketch



import lizard
import pathlib
import unittest


class TestCCN(unittest.TestCase):

    def setUp(self) -> None:
        """
        初始化 lizard
        """
        self.debug = False
        self.results = []
        path = pathlib.Path(__file__).parent.parent
        for file in path.rglob("*.py"):
            if self.debug:
                print(file.absolute())
            result = lizard.analyze_file(str(file.absolute()))
            self.results.append(result)
            if self.debug:
                print("result:", result.__dict__)
                result = result.__dict__
                for func in result['function_list']:
                    print("function:", func.name)
                    print(func.__dict__)
                print("=" * 30)

    def test_nloc(self):
        """
        不带注释的代码行检查
        """
        for result in self.results:
            for func in result.function_list:
                self.assertLess(func.nloc, 100, "{}:不带注释的代码行应该小于100,但是当前是{}".format(func.name, func.nloc))
            self.assertLess(result.nloc, 100, "{}:不带注释的代码行应该小于100,但是当前是{}".format(result.filename, result.nloc))

    def test_ccn(self):
        """
        圈复杂度检查
        """
        for result in self.results:
            for func in result.function_list:
                self.assertLess(func.cyclomatic_complexity, 20,
                                "{}:圈复杂度应当小于20,但是当前是{}".format(func.name, func.cyclomatic_complexity))
            self.assertLess(result.average_cyclomatic_complexity, 10,
                            "{}:圈复杂度应当小于10,但是当前是{}".format(result.filename, result.average_cyclomatic_complexity))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: 
# @time: 
# @Software: PyCharm


import math
import numpy as np
from scipy.spatial.distance import cdist, pdist, euclidean

from snowland.graphics.core.computational_geometry_2d import Point2D
from snowland.graphics.core.computational_geometry_3d import Point3D
from snowland.graphics.core.computational_geometry_base import Vector, Vector3
from snowland.graphics.core.analytic_geometry_base import Line, Surface

npa = np.array
npl = np.linalg

__all__ = [
    'Line2D'
]


class Line2D(Line):
    def __init__(self, a=None, b=0, c=0, p1: Point2D = None, p2: Point2D = None, *args, **kwargs):
        if a is not None and b is not None and c is not None:
            # ax+by+c=0方式定义的线
            self.a, self.b, self.c = a, b, c
        elif p1 is not None and p2 is not None:
            # 两点式定义的直线
            if isinstance(p1, Point2D) or isinstance(p2, Point2D):
                raise ValueError("p1, p2应为二维点Point2D的子类")
            self.a, self.b, self.c = p2.y - p1.y, p1.x - p2.x, -a * p1.x - b * p1.y
        else:
            # TODO: 其他方式确定的直线
            pass

    def __distance_point(self, point):
        """
        计算点到直线距离
        :param point: 点
        :return:
        """
        p = Point2D(point)
        return np.fabs(self.a * p.x + self.b * p.y + self.c) / np.sqrt(self.a ** 2 + self.b ** 2)

    def distance_points(self, points):
        """
        计算多点到直线的距离
        :param points: 可以是np.ndarray, 也可以是list(Point2D)
        :return:
        """
        # TODO: 这里可以用矩阵进行运算的，以后补上
        if isinstance(points, list):
            return npa([self.__distance_point(each) for each in points])
        elif isinstance(points, np.ndarray):
            if len(points.shape) == 2:
                return npa([self.__distance_point(each) for each in points])
            else:
                return self.__distance_point(points)
        else:
            raise ValueError('points 参数错误')

    def intersection(self, other, eps=1e-8):
        """
        直线交点
        :param other:
        :param eps:
        :return:
        """
        if not isinstance(other, Line2D):
            pass
        if -eps < (self.a - other.a) < eps and -eps < self.b - other.b < eps:
            return None  # 不相交的两条直线
        else:
            return Point2D([-(self.b * other.c - self.c * other.b) / (other.a * self.b - self.a * other.b), \
                            (self.a * other.c - self.c * other.a) / (other.a * self.b - self.a * other.b)])




# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: _ploygon.py
# @time: 2019/8/12 20:48
# @Software: PyCharm

from snowland.graphics.core.computational_geometry_2d import Point2D, Polygon, PolygonWithoutHoles, LineString2D, ConvexPolygon, Rectangle
from snowland.graphics.core.computational_geometry_base import Vector
from snowland.graphics.core.analytic_geometry_2d import Line2D
import numpy as np

npl = np.linalg
npa = np.array
npr = np.random

def __in_polygon(p: Point2D, polygon):
    """
    判断点是否在多边形内（不包含边上）
    :param point:
    :param polygon:
    :return:
    """
    for point in polygon.p:
        if p == point:  # 使用重载的等于方法
            return True

    for hole in polygon.holes:
        for point in hole:
            if p == point:  # 使用重载的等于方法
                return True
    count = 0
    flag = False
    for i, point in enumerate(polygon.p[:, :]):  # 注意，这个i取0的时候，i-1值是-1，所以这个是循环到所有的边
        if (point[1] <= p[1] < polygon.p[i - 1, 1] or polygon.p[i - 1, 1] <= p[1] < point[1]) and \
                (p[0] < (polygon.p[i - 1, 0] - point[0]) * (p[1] - point[1]) / (polygon.p[i - 1, 1] - point[1]) + point[
                    1]):
            flag = not flag
            count += 1

    for hole in polygon.holes:
        polygon_temp = PolygonWithoutHoles(hole)
        for i, point in enumerate(polygon_temp.p[:, :]):  # 注意，这个i取0的时候，i-1值是-1，所以这个是循环到所有的边
            if (point[1] <= p[1] < polygon_temp.p[i - 1, 1] or polygon_temp.p[i - 1, 1] <= p[1] < point[1]) and \
                    (p[0] < (polygon_temp.p[i - 1, 0] - point[0]) *
                     (p[1] - point[1]) / (polygon_temp.p[i - 1, 1] - point[1]) + point[1]):
                flag = not flag
                count += 1
    return flag


def in_polygon(point, polygon: Polygon):
    """
    判断点是否在多边形内
    :param point:
    :param polygon:
    :return:
    """
    if isinstance(point, Point2D):
        return __in_polygon(point, polygon)
    if isinstance(point, np.ndarray) and len(point.shape) == 1:
        return __in_polygon(Point2D(point), polygon)
    else:
        # 多个点判断， 返回多个值
        if isinstance(point[0], Point2D):
            return npa([__in_polygon(p, polygon) for p in point])
        if isinstance(point[0], np.ndarray) and len(point.shape) == 1:
            return npa([__in_polygon(Point2D(p), polygon) for p in point])
        raise ValueError('error')


# 使用Graham扫描法计算凸包
# 网上的代码好多运行效果并不好
# 算法参见《算法导论》第三版 第605页
# https://blog.csdn.net/john_bian/article/details/85221039


def get_bottom_point(points):
    """
    返回points中纵坐标最小的点的索引，如果有多个纵坐标最小的点则返回其中横坐标最小的那个
    :param points:
    :return:
    """
    min_index = 0
    n = len(points)
    for i in range(0, n):
        if points[i][1] < points[min_index][1] or (
                points[i][1] == points[min_index][1] and points[i][0] < points[min_index][0]):
            min_index = i
    return min_index


def sort_polar_angle_cos(points, center_point: Point2D):
    """
    按照与中心点的极角进行排序，使用的是余弦的方法
    :param points: 需要排序的点
    :param center_point: 中心点
    :return:
    """
    n = len(points)
    cos_value = []
    rank = []
    norm_list = []
    for i in range(0, n):
        point_ = points[i]
        point = [point_[0] - center_point[0], point_[1] - center_point[1]]
        rank.append(i)
        norm_value = npl.norm(point)
        norm_list.append(norm_value)
        if norm_value == 0:
            cos_value.append(1)
        else:
            cos_value.append(point[0] / norm_value)

    for i in range(0, n - 1):
        index = i + 1
        while index > 0:
            if cos_value[index] > cos_value[index - 1] or (
                    cos_value[index] == cos_value[index - 1] and norm_list[index] > norm_list[index - 1]):
                temp = cos_value[index]
                temp_rank = rank[index]
                temp_norm = norm_list[index]
                cos_value[index] = cos_value[index - 1]
                rank[index] = rank[index - 1]
                norm_list[index] = norm_list[index - 1]
                cos_value[index - 1] = temp
                rank[index - 1] = temp_rank
                norm_list[index - 1] = temp_norm
                index = index - 1
            else:
                break
    sorted_points = []
    for i in rank:
        sorted_points.append(points[i])

    return sorted_points


def vector_angle(vector):
    """
    返回一个向量与向量 [1, 0]之间的夹角， 这个夹角是指从[1, 0]沿逆时针方向旋转多少度能到达这个向量
    :param vector:
    :return:
    """
    norm_ = npl.norm(vector)
    if norm_ == 0:
        return 0

    angle = np.arccos(vector[0] / norm_)
    if vector[1] >= 0:
        return angle
    else:
        return 2 * np.pi - angle


def coss_multi(v1, v2):
    """
    计算两个向量的叉乘
    :param v1:
    :param v2:
    :return:
    """
    return v1[0] * v2[1] - v1[1] * v2[0]


def graham_scan(points):
    # print("Graham扫描法计算凸包")
    bottom_index = get_bottom_point(points)
    bottom_point = points.pop(bottom_index)
    sorted_points = sort_polar_angle_cos(points, bottom_point)

    m = len(sorted_points)
    if m < 2:
        print("点的数量过少，无法构成凸包")
        return

    stack = []
    stack.append(bottom_point)
    stack.append(sorted_points[0])
    stack.append(sorted_points[1])

    for i in range(2, m):
        length = len(stack)
        top = stack[length - 1]
        next_top = stack[length - 2]
        v1 = [sorted_points[i][0] - next_top[0], sorted_points[i][1] - next_top[1]]
        v2 = [top[0] - next_top[0], top[1] - next_top[1]]

        while coss_multi(v1, v2) >= 0:
            stack.pop()
            length = len(stack)
            top = stack[length - 1]
            next_top = stack[length - 2]
            v1 = [sorted_points[i][0] - next_top[0], sorted_points[i][1] - next_top[1]]
            v2 = [top[0] - next_top[0], top[1] - next_top[1]]

        stack.append(sorted_points[i])

    return ConvexPolygon(stack)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: __init__.py .py
# @time: 2020/9/6 23:48
# @Software: PyCharm



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.snowland.ltd
# @file: ing_to_oilpaint.py
# @time: 2018/8/12 2:15
# @Software: PyCharm

from copy import deepcopy

import numpy as np


# https://www.cnblogs.com/lonelyxmas/p/ 8564738.html
def oilpaint(img, intensity=5):
    if np.max(img) > 1.1:
        img = img / 255
    tempValues = deepcopy(img)
    w, h, _ = img.shape
    rand = np.random.randint(0, intensity, (w, h))
    X, Y = np.meshgrid(range(h), range(w))
    X, Y = X + rand, Y + rand
    X[X >= h] = h - 1
    Y[Y >= w] = w - 1
    tempValues[:, :, :] = img[Y, X, :]
    return tempValues



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: 
# @file: img_to_handpainting.py
# @time: 2018/8/8 21:42
# @Software: PyCharm

# 代码修改自 https://github.com/Leungtamir/Image-Freehand/blob/master/image.py

import numpy as np
from skimage.color import rgb2grey


def freehand(img, depth=10., el=np.pi / 2.2, az=np.pi / 4):
    """
    手绘风格图像生成
    :param img:
    :param depth: 深度，取值在0-100
    :param el: 光源的俯视角度，弧度值
    :param az: 光源的方位角度，弧度值
    :return:
    """
    img = rgb2grey(img)
    img = img * 255 if np.max(img) <= 1.1 else img
    grad = np.gradient(img)  # 取图像灰度的梯度值
    grad_x, grad_y = grad  # 分别取横纵图像梯度值
    grad_x = grad_x * depth / 100.
    grad_y = grad_y * depth / 100.
    A = np.sqrt(grad_x ** 2 + grad_y ** 2 + 1.)
    uni_x = grad_x / A
    uni_y = grad_y / A
    uni_z = 1. / A

    dx = np.cos(el) * np.cos(az)  # 光源对x轴的影响
    dy = np.cos(el) * np.sin(az)  # 光源对y轴的影响
    dz = np.sin(el)  # 光源对z轴的影响

    gd = (dx * uni_x + dy * uni_y + dz * uni_z)  # 光源归一化
    gd = gd.clip(0, 1)  # 避免数据越界，将生成的灰度值裁剪至0-1之间
    return gd



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: 
# @time: 
# @Software: PyCharm


from snowland.exceptions._exception import *



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: 
# @time: 
# @Software: PyCharm


import numpy as np

from snowland.graphics.core.computational_geometry_2d import Point2D
from snowland.graphics.core.computational_geometry_3d import Point3D
from snowland.graphics.core.computational_geometry_base import Vector, Vector3
from snowland.graphics.core.analytic_geometry_base import Line, Surface

npa = np.array
npl = np.linalg
__all__ = [
    'Line3D',
    'Plane3D'
]


class Line3D(Line):
    def __init__(self, x0=0, y0=0, z0=0, p: Point3D = None, a=None, b=None, c=None, p1: Point3D = None,
                 p2: Point3D = None, plane1=None, plane2=None, *args, **kwargs):
        if a is not None and b is not None and c is not None:
            # (x-x0)/a = (y-y0)/b = (z-z0)/c
            self.a, self.b, self.c = a, b, c
            if p is None:
                # TODO: 判断x0, y0, z0是不是为空
                p = x0, y0, z0
            self.p = Point3D(p)
        elif p1 is not None and p2 is not None:
            # 两点式定义的直线
            if isinstance(p1, Point3D) or isinstance(p2, Point3D):
                raise ValueError("p1, p2应为三维点Point3D的子类")
            self.a, self.b, self.c = p2.x - p1.x, p2.y - p1.y, p2.z - p1.z
            if p is None:
                # TODO: 判断x0, y0, z0是不是为空
                p = x0, y0, z0
            self.p = Point3D(p)
        else:

            if isinstance(plane1, Plane3D) and isinstance(plane2, Plane3D):
                # TODO: 两平面夹线的处理， 此方法有局限 A.Star<astar@snowland.ltd> 2020-9-7
                # 参考资料： http://www.360doc.com/content/15/1222/10/17164483_522220811.shtml
                a1, b1, c1, d1 = plane1.a, plane1.b, plane1.c, plane1.d
                a2, b2, c2, d2 = plane2.a, plane2.b, plane2.c, plane2.d
                self.a, self.b, self.c = npl.det([[b1, c1], [b2, c2]]), npl.det([[a1, c1], [a2, c2]]), npl.det([[a1, b1], [a2, b2]])
                self.p = Point3D((b1*d2-b2*d1)/(a1*b2-a2*b1), (a1*d2-a2*d1)/(a2*b1-a1*b2), 0)

    def feature_vector(self):
        """
        特征向量
        :return: 
        """
        return Vector3([self.a, self.b, self.c])


class Plane3D(Surface):
    def __init__(self, a=None, b=None, c=None, d=None, v=None, p0=None, p1=None, p2=None, p3=None, *args, **kwargs):
        if a is not None and b is not None and c is not None and d is not None:
            # ax + by + cz + d = 0
            self.a, self.b, self.c, self._d = a, b, c, d
        elif v is not None and p0 is not None:
            if isinstance(v, Vector3):
                self.a, self.b, self.c = v.x[0], v.x[1], v.x[2]
            elif isinstance(v, (list, np.ndarray)):
                self.a, self.b, self.c = v[0], v[1], v[2]
                v = Vector3(v)
            else:
                raise ValueError('类型不匹配')

            if isinstance(p0, Point3D):
                p = p0.p
            elif isinstance(p0, (list, np.ndarray)):
                p = Point3D(p0)
            else:
                raise ValueError('类型不匹配')
            self.d = - v.x @ p.p
        else:
            # TODO: 三点式
            pass
        pass

    def normal_vector(self):
        """
        特征向量
        :return:
        """
        return Vector3([self.a, self.b, self.c])

    def is_vector_parallel(self, vector: (Vector3, np.ndarray, list)):
        """
        判断直线是否与平面是否平行
        :param plane1:
        :param plane2:
        :return:
        """
        return self.normal_vector().is_vertical(vector)

    def is_vector_vertical(self, vector: (Vector3, np.ndarray, list)):
        """
        判断向量是否与平面是否垂直
        :param vector:
        :return:
        """
        return self.normal_vector().is_parallel(vector)

    def angle_plane(self, plane):
        """
        平面间夹角
        :param plane:
        :return:
        """
        return self.normal_vector().angle(plane.normal_vector())



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: qgis_utils.py
# @time: 2022/01/06 20:52
# @Software: PyCharm


from functools import reduce

import numpy as np
from qgis.core import QgsGeometry, QgsPoint

from snowland.qgis_tool.haversine import length_metres

npa = np.array
npl = np.linalg


__all__ = [
    'points',
    'pointsXY',
    'vector',
    'vector_start',
    'vector_end',
    'point_to_ndarray',
    'start_pointXY',
    'end_pointXY'
]


def points(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y(), v.z()) for v in vertices])


def pointsXY(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y()) for v in vertices])


def vector(geometry: QgsGeometry):
    """
    起点到终点的向量
    """
    start = geometry.vertexAt(0)
    line_nodes = geometry.asPolyline()
    end = line_nodes[-1]
    return npa([end.x() - start.x(), end.y() - start.y()])


def vector_start(geometry: QgsGeometry):
    """
    第一条折线的向量
    """
    start = geometry.vertexAt(0)
    end = geometry.vertexAt(1)
    return npa([end.x() - start.x(), end.y() - start.y()])


def vector_end(geometry: QgsGeometry):
    """
    倒数第一条折线的向量
    """
    line_nodes = geometry.asPolyline()
    start = line_nodes[-2]
    end = line_nodes[-1]
    return npa([end.x() - start.x(), end.y() - start.y()])


def thinning_by_step_method(points, step=2):
    """
    步长法
    :param points:
    :return:
    """
    result = [points[0]]
    delta_length = length_metres(points[:-1], points[1:])
    delta = 0
    for i, d in enumerate(delta_length, start=1):
        if delta + d > step:
            result.append(points[i])
            delta = 0
        else:
            delta += d
    if delta != 0:
        result.append(points[-1])
    return reduce(lambda a, b: np.vstack((a, b)), result)


def douglas_peuker(points, d=5e-7):
    """
    Douglas Peuker
    :param points:
    :return:
    """
    if len(points) < 2:
        return points
    p1 = points[0]
    p2 = points[-1]
    line = npa([p2[1] - p1[1], p1[0] - p2[0], p2[0] * p1[1] - p1[0] * p2[1]])

    dist = np.abs(points[:, 0] * line[0] + points[:, 1] * line[1] + line[2]) / npl.norm(line[:2])
    d_ind = np.argmax(dist)
    if dist[d_ind] < d:
        return npa((p1, p2))
    else:
        return np.vstack((douglas_peuker(points[:d_ind + 1], d), douglas_peuker(points[d_ind:], d)[1:]))


def vertical_distance_limit(points, d=5e-7):
    """
    垂距限值法
    :param points:
    :return:
    """
    p0 = points[0]
    if len(points) < 3:
        return points
    result = [p0]

    for p1, p2 in zip(points[1:-1, :], points[2:, :]):
        line = npa([p2[1] - p0[1], p0[0] - p2[0], p2[0] * p0[1] - p0[0] * p2[1]])
        dist = p1[0] * line[0] + p1[1] * line[1] + line[2]
        print(dist)
        if dist > d:
            result.append(p1)
            p0 = p1
    result.append(points[-1])
    return reduce(lambda a, b: np.vstack((a, b)), result)


def start_pointXY(geometry: QgsGeometry):
    v = geometry.vertexAt(0)
    return npa([v.x(), v.y()])


def end_pointXY(geometry: QgsGeometry):
    v = geometry.asPolyline()[-1]
    return npa([v.x(), v.y()])


def point_to_ndarray(p: QgsPoint):
    return npa([p.x(), p.y()])



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file:
# @time:
# @Software: PyCharm

from scipy.spatial.distance import pdist, cdist
import numpy as np

from snowland.graphics.core.computational_geometry_base import Point, LineString, Shape

npa = np.array
npm = np.mat


__all__ = [
    'Circle',
    'ConvexPolygon',
    'Diamond',
    'Ellipse',
    'LineSegment2D',
    'LineString2D',
    'Point2D',
    'Polygon',
    'PolygonWithoutHoles',
    'Rectangle',
    'Square',
    'Triangle'
]

class Point2D(Point):
    def __init__(self, p=None, x=None, y=None):
        if x is not None and y is not None:
            p = (x, y)
        assert p is not None
        super(Point2D, self).__init__(p)

    @property
    def x(self):
        return self.p[0]

    @x.setter
    def set_x(self, px):
        self.p[0] = px

    @property
    def y(self):
        return self.p[1]

    @y.setter
    def set_y(self, py):
        self.p[1] = py


class LineString2D(LineString):
    def __init__(self, X=None):
        super(LineString2D, self).__init__(X=X)
        assert self.X.shape[0] is 2


class LineSegment2D(LineString2D):
    def __init__(self, X=None, p1=None, p2=None):
        if p1 is not None and p2 is not None:
            if isinstance(p1, Point):
                p1 = p1[:2]
            if isinstance(p2, Point):
                p2 = p2[:2]
            X = np.vstack((p1, p2))
        assert X is not None
        super(LineSegment2D, self).__init__(X)

    def length(self, metric='euclidean', *args, **kwargs):
        return np.sum(pdist(self.X, metric, *args, **kwargs))


class Polygon(Shape):
    def __init__(self, p=None, holes=None):
        self.p = np.zeros((0, 2))
        if p:
            if isinstance(p, np.ndarray):
                self.p = p
            else:
                if isinstance(p[0], Point2D):
                    for point in p:
                        self.p = np.vstack((self.p, [point.p]))
                else:
                    for point in p:
                        self.p = np.vstack((self.p, [point]))
            if np.all(self.p[0, :] == self.p[-1, :]):
                self.p = self.p[:-1, :]
        self.holes = []
        if holes:
            for hole in holes:
                if isinstance(hole, np.ndarray):
                    h = hole
                else:
                    h = np.zeros((0, 2))
                    if isinstance(hole[0], Point2D):
                        for point in h:
                            h = np.vstack((h, [point.p]))
                    else:
                        for point in p:
                            h = np.vstack((h, [point]))
                if h[0, :] == h[-1, :]:
                    h = h[:-1, :]
                self.holes.append(h)

    def area(self):
        p = PolygonWithoutHoles(self.p)
        area_p = p.area()
        area_holes = [PolygonWithoutHoles(h).area() for h in self.holes]
        return area_p - np.sum(area_holes)

    def girth(self):
        p = PolygonWithoutHoles(self.p)
        girth_p = p.girth()
        girth_holes = [PolygonWithoutHoles(h).girth() for h in self.holes]
        return girth_p - np.sum(girth_holes)

    def without_holes(self):
        return self.holes is None or not self.holes

    def is_convex(self):
        if not self.without_holes():
            return False
        # TODO, check


class PolygonWithoutHoles(Polygon):
    def __init__(self, p):
        super(PolygonWithoutHoles, self).__init__(p, holes=None)

    def area(self):
        """
        面积
        :return:
        """
        # TODO 计算面积

    def girth(self):
        """
        周长
        :return:
        """
        p_ring = np.vstack((self.p, self.p[:1, :]))
        return LineString(p_ring).length()


class ConvexPolygon(PolygonWithoutHoles):

    def area(self):
        # 面积
        # 参考 https://blog.csdn.net/cymy001/article/details/81366033
        points = self.p
        area = 0
        if len(points) < 3:
            raise Exception("error")

        for i, p1 in enumerate(points[:, :]):  # 当i = 0时首尾计算面积
            p2 = points[i - 1]
            triArea = (p1[0] * p2[1] - p2[0] * p1[1]) / 2
            # print(triArea)
            area += triArea
        return abs(area)


class Triangle(ConvexPolygon):
    """
    三角形
    """
    pass


class Rectangle(ConvexPolygon):
    pass


class Diamond(ConvexPolygon):
    def __init__(self, p):
        super(PolygonWithoutHoles, self).__init__(p, holes=None)
        m, n = self.p.shape
        dist = [pdist(self.p[ind:ind + 2, :]) for ind in range(m - 1)]
        d = cdist(self.p[:1, :], self.p[-1:, :])
        assert np.all(dist == d)


class Square(Rectangle, Diamond):
    """
    正方形
    """

    def __init__(self, p, eps=1e-8, *args, **kwargs):
        super(Square, self).__init__(p, *args, **kwargs)
        # TODO 判断方形


class Ellipse(Shape):
    def __init__(self, centre, a=0, b=0):
        self.centre = Point2D(centre)
        self.a = a
        self.b = b

    def area(self):
        """
        面积
        :return:
        """
        return self.a * self.b * np.pi

    def girth(self):
        """
        周长
        :return:
        """
        return 4 * self.a * self.b * np.pi

    def is_circle(self, eps=1e-8):
        return -eps <= (self.a - self.b) <= eps


class Circle(Ellipse):
    def __init__(self, centre, r=0):
        super(Circle, self).__init__(centre, r, r)

    @property
    def r(self):
        return self.a

    @r.setter
    def set_r(self, r):
        self.a = self.b = r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: qgis_tool.py
# @time: 2022/01/06 23:48
# @Software: PyCharm

import sys
import unittest
from astartool.project import std_logging
from astartool.number import equals_zero
from qgis.core import QgsDistanceArea, QgsUnitTypes, QgsPointXY
import pathlib

this_file = pathlib.Path(__file__)
path_pathlib = this_file.parent.parent.parent
path = str(path_pathlib)
print("path:", path)

if path not in sys.path:
    sys.path.append(path)

from snowland.qgis_tool import haversine


class TestHaversine(unittest.TestCase):

    @classmethod
    @std_logging()
    def setup_class(cls):
        pass

    @classmethod
    @std_logging()
    def teardown_class(cls):
        pass

    @std_logging()
    def setup_method(self, method):
        pass

    @std_logging()
    def teardown_method(self, method):
        pass

    @std_logging()
    def setUp(self):
        Ellipsoid = (haversine.EARTH_REDIUS * 1000, haversine.EARTH_REDIUS * 1000)
        disClass = QgsDistanceArea()
        disClass.setEllipsoid(Ellipsoid[0], Ellipsoid[1])

        self.module = haversine
        self.disClass = disClass

    @std_logging()
    def tearDown(self):
        pass

    def test_1(self):
        """
        经度上加一个值
        """
        p1 = [118, 40]
        p2 = [118.1, 40]
        p1_pointxy = QgsPointXY(*p1)
        p2_pointxy = QgsPointXY(*p2)
        line = self.disClass.measureLine(p1_pointxy, p2_pointxy)
        qgis_meters = self.disClass.convertLengthMeasurement(line, QgsUnitTypes.DistanceMeters)
        haversine_meters = self.module.haversine_metres(*p1_pointxy, *p2_pointxy)
        self.assertTrue(equals_zero(qgis_meters - haversine_meters, 1e-4))  # 毫米量级

    def test_2(self):
        """
        纬度上加一个值
        """
        p1 = [118, 40]
        p2 = [118, 40.1]
        p1_pointxy = QgsPointXY(*p1)
        p2_pointxy = QgsPointXY(*p2)
        line = self.disClass.measureLine(p1_pointxy, p2_pointxy)
        qgis_meters = self.disClass.convertLengthMeasurement(line, QgsUnitTypes.DistanceMeters)
        haversine_meters = self.module.haversine_metres(*p1_pointxy, *p2_pointxy)
        self.assertTrue(equals_zero(qgis_meters - haversine_meters, 1e-4))  # 毫米量级

    def test_3(self):
        """
        测试相同的点
        """
        p1 = [0, 40]
        p2 = [0, 40]
        p1_pointxy = QgsPointXY(*p1)
        p2_pointxy = QgsPointXY(*p2)
        line = self.disClass.measureLine(p1_pointxy, p2_pointxy)
        qgis_meters = self.disClass.convertLengthMeasurement(line, QgsUnitTypes.DistanceMeters)
        haversine_meters = self.module.haversine_metres(*p1_pointxy, *p2_pointxy)
        self.assertTrue(equals_zero(qgis_meters - haversine_meters, 1e-4))  # 毫米量级

    def test_4(self):
        """
        跨南北半球
        """
        p1 = [-0.1, 0]
        p2 = [0.1, 0]
        p1_pointxy = QgsPointXY(*p1)
        p2_pointxy = QgsPointXY(*p2)
        line = self.disClass.measureLine(p1_pointxy, p2_pointxy)
        qgis_meters = self.disClass.convertLengthMeasurement(line, QgsUnitTypes.DistanceMeters)
        haversine_meters = self.module.haversine_metres(*p1_pointxy, *p2_pointxy)
        self.assertTrue(equals_zero(qgis_meters - haversine_meters, 1e-4))  # 毫米量级

    def test_5(self):
        """
        极大
        """
        p1 = [90, 0]
        p2 = [-90, 180]
        p1_pointxy = QgsPointXY(*p1)
        p2_pointxy = QgsPointXY(*p2)
        line = self.disClass.measureLine(p1_pointxy, p2_pointxy)
        qgis_meters = self.disClass.convertLengthMeasurement(line, QgsUnitTypes.DistanceMeters)
        haversine_meters = self.module.haversine_metres(*p1_pointxy, *p2_pointxy)
        print(qgis_meters, haversine_meters)
        self.assertTrue(equals_zero(qgis_meters - haversine_meters, 1e-4))  # 毫米量级


# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: solid_geometry_3d.py
# @time: 2019/8/17 17:13
# @Software: PyCharm

from abc import abstractmethod

from scipy.spatial.distance import pdist
from skimage.draw.draw import *

from snowland.graphics.core.computational_geometry_base import Point, LineString, Shape, Stereograph
from snowland.graphics.core.computational_geometry_base import UNITVECTORZ

npa = np.array
npm = np.mat

__all__ = [
    'Circle3D',
    'ConvexPolygon3D',
    'Cylinder',
    'Ellipse3D',
    'LineSegment3D',
    'LineString3D',
    'Point3D',
    'Polygon3D',
    'PolygonWithoutHoles3D',
    'Sphere',
    'Square3D',
    'StraightCone',
    'StraightCylinder',
    'StraightEllipsoid',
    'StraightPrism',
    'Triangle3D',
    'Vertebrae'
]


class Point3D(Point):
    def __init__(self, p=None, x=None, y=None, z=None):
        if x is not None and y is not None and z is not None:
            p = npa((x, y, z))
        assert p is not None
        super(Point3D, self).__init__(p)

    @property
    def x(self):
        return self.p[0]

    @x.setter
    def set_x(self, px):
        self.p[0] = px

    @property
    def y(self):
        return self.p[1]

    @y.setter
    def set_y(self, py):
        self.p[1] = py

    @property
    def z(self):
        return self.p[0]

    @z.setter
    def set_z(self, px):
        self.p[0] = px


class LineString3D(LineString):
    def __init__(self, X=None):
        super(LineString3D, self).__init__(X=X)
        assert self.X.shape[0] is 3


class LineSegment3D(LineString3D):
    def __init__(self, X=None, p1=None, p2=None):
        if p1 is not None and p2 is not None:
            if isinstance(p1, Point):
                p1 = p1[:3]
            if isinstance(p2, Point):
                p2 = p2[:3]
            X = np.vstack((p1, p2))
        super(LineSegment3D, self).__init__(X)

    def length(self, metric='euclidean', *args, **kwargs):
        return np.sum(pdist(self.X, metric, *args, **kwargs))


class Shape3D(Shape):
    """
    三维面
    """
    pass


class Polygon3D(Shape3D):
    def __init__(self, p=None, holes=None):
        # TODO:
        super(Polygon3D, self).__init__()
        self.p = p
        self.holes = holes

    def area(self):
        # TODO:
        pass

    def girth(self):
        # TODO
        pass

    def without_holes(self):
        return self.holes is None or not self.holes

    def is_convex(self):
        if not self.without_holes():
            return False
        # TODO, check


class PolygonWithoutHoles3D(Polygon3D):
    def __init__(self, p):
        super(PolygonWithoutHoles3D, self).__init__(p, holes=None)

    def area(self):
        """
        面积
        :return:
        """
        # TODO 计算面积

    def girth(self):
        """
        周长
        :return:
        """
        p_ring = np.vstack((self.p, self.p[:1, :]))
        return LineString3D(p_ring).length()


class ConvexPolygon3D(PolygonWithoutHoles3D):

    def area(self):
        # 面积
        # 参考 https://blog.csdn.net/cymy001/article/details/81366033
        points = self.p
        area = 0
        if len(points) < 3:
            raise Exception("error")

        for i in range(0, len(points) - 1):
            p1 = points[i]
            p2 = points[i + 1]

            triArea = (p1[0] * p2[1] - p2[0] * p1[1]) / 2
            # print(triArea)
            area += triArea

        fn = (points[-1][0] * points[0][1] - points[0][0] * points[-1][1]) / 2
        # print(fn)
        return abs(area + fn)


class Triangle3D(ConvexPolygon3D):
    """
    三角形
    """
    pass


class Square3D(ConvexPolygon3D):
    """
    正方形
    """
    pass


class Ellipse3D(Shape3D):
    def __init__(self, centre, a=0, b=0, v=UNITVECTORZ):
        self.centre = Point3D(centre)
        self.a = a
        self.b = b
        self.v = v

    def area(self):
        """
        面积
        :return:
        """
        return self.a * self.b * np.pi

    @abstractmethod
    def girth(self):
        """
        周长
        :return:
        """
        # TODO: 椭圆周长数值解

    def is_circle(self, eps=1e-8):
        return np.fabs(self.a - self.b) <= eps


class Circle3D(Ellipse3D):
    def __init__(self, centre, r=0, v=UNITVECTORZ):
        super(Circle3D, self).__init__(centre, r, r, v)

    @property
    def r(self):
        return self.a

    @r.setter
    def set_r(self, ra):
        self.a = self.b = ra

    def girth(self):
        """
        周长
        :return:
        """
        return 2 * self.r * np.pi


class Vertebrae(Stereograph):
    """
    椎体
    """

    def __init__(self, underside: Shape3D = None, h=0, v=UNITVECTORZ, t=UNITVECTORZ):
        self.underside = underside
        self.h = h
        self.v = v  # 底面法向量
        self.t = t  # 中线向量

    @abstractmethod
    def surface_area(self):
        """
        表面积
        :return:
        """
        pass

    def volume(self):
        """
        体积
        :return:
        """
        return self.underside.area() * self.h / 3


class StraightCone(Vertebrae):
    # 圆锥
    def __init__(self, underside: Circle3D = None, h=0, v=UNITVECTORZ, centre=None, r=0):
        """
        :param underside:
        :param centre: 底面圆心
        :param r: 底面半径
        :param h: 圆锥高度
        :param v: 方向向量
        """
        if underside is None:
            underside = Circle3D(centre, r, v)

        super(StraightCone, self).__init__(underside=underside, h=h, v=v, t=v)

    def surface_area(self):
        """
        表面积
        :return:
        """
        return self.underside.area() + self.underside.r * np.sqrt(self.h * self.h + self.underside.r * self.underside.r)


class Cylinder(Stereograph):
    """
    柱体
    """

    def __init__(self, underside: Shape3D = None, h=0, v=UNITVECTORZ, t=UNITVECTORZ):
        self.underside = underside
        self.h = h
        self.v = v
        self.t = t

    def surface_area(self):
        """
        表面积
        :return:
        """
        return 2 * self.underside.area() + self.underside.girth() * self.h

    def volume(self):
        """
        体积
        :return:
        """
        return self.underside.area() * self.h


class StraightCylinder(Cylinder):
    def __init__(self, underside: Circle3D = None, h=0, v=UNITVECTORZ, centre=None, r=0):
        """
        :param centre: 底面圆心
        :param r: 底面半径
        :param h: 圆柱高度
        :param v: 方向向量
        """
        if underside is None:
            underside = Circle3D(centre, r)

        super(StraightCylinder, self).__init__(underside, h, v, t=v)


class StraightPrism(Cylinder):
    def __init__(self, underside: ConvexPolygon3D = None, h=0, v=UNITVECTORZ, p=None):
        """
        :param underside: 底面
        :param h: 圆柱高度
        :param v: 方向向量
        """
        if underside is None:
            if p is not None:
                underside = Polygon3D(p)
            else:
                raise ValueError('初始化异常')

        super(StraightPrism, self).__init__(underside, h, v, t=v)


class StraightEllipsoid(Stereograph):
    def __init__(self, a=1, b=1, c=1, p=(0, 0, 0)):
        # (x-x0)**2 / a**2 + (y-y0)**2 / b**2+(z-z0)**2 / c**2=1
        self.p = Point3D(p)
        self.a = a
        self.b = b
        self.c = c

    def volume(self):
        return 4 * np.pi * self.a * self.b * self.c / 3

    def surface_area(self):
        return 4 * np.pi * (self.a * self.b + self.b * self.c + self.c * self.a) / 3


class Sphere(StraightEllipsoid):
    # 球体
    def __init__(self, r=1, p=(0, 0, 0)):
        super(Sphere, self).__init__(r, r, r, p)

    @property
    def r(self):
        return self.a



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: __init__.py
# @time: 2019/8/12 20:47
# @Software: PyCharm


from snowland.graphics.solid_geometry._ploygon import in_polygon



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : 河北雪域网络科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: 
# @file: setup.py
# @time: 2018/7/11 15:13
# @Software: PyCharm


from setuptools import setup, find_packages
from astartool.setuptool import load_install_requires
from snowland import __version__

setup(
    name='snowland-image',
    version=__version__,
    description=(
        'toolkit for image'
    ),
    long_description=open('README.rst', encoding='utf-8').read(),
    author='A.Star',
    author_email='astar@snowland.ltd',
    maintainer='A.Star',
    maintainer_email='astar@snowland.ltd',
    license='BSD License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/AAFun/scikit-snowland',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=load_install_requires()
)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: solid_geometry_base.py
# @time: 2019/8/17 0:04
# @Software: PyCharm


from abc import ABCMeta, abstractmethod
from numbers import Number

from scipy.spatial.distance import pdist
import numpy as np
from astartool.number import equals_zero_all

npa = np.array
npl = np.linalg

__all__ = [
    'Graphic',
    'LineString',
    'Point',
    'Shape',
    'Stereographic',
    'UnitVector',
    'UnitVector3',
    'UNITVECTORX',
    'UNITVECTORY',
    'UNITVECTORZ',
    'Vector',
    'Vector2',
    'Vector3',
]


class Vector(object):
    def __init__(self, x=None, dim=None, start=None, end=None):
        if isinstance(x, Vector):
            x = x.x
        elif dim is not None:
            x = np.zeros(dim)
        elif x is not None:
            # TODO: 判断x是一维的
            x = npa(x)
        elif start is not None and end is not None:
            x = Point(end) - Point(start)

        self.x = x

    def __mul__(self, other):
        v2 = Vector(other)
        return v2.x * self.x

    def __add__(self, other):
        v2 = Vector(other)
        return Vector(v2.x + self.x)

    def __sub__(self, other):
        v2 = Vector(other)
        return Vector(self.x - v2.x)

    def __str__(self):
        itor = [str(_) for _ in self.x]
        return "Vector <" + ', '.join(itor) + ">"

    def length(self):
        """
        向量模长
        :return:
        """
        return npl.norm(self.x)

    def dot_product(self, other):
        v2 = Vector(other)
        return self.x.dot(v2.x)

    def cross_product(self, other):
        # TODO: 向量积
        raise NotImplemented

    def cos_angle(self, other):
        v2 = Vector(other)
        return self.dot_product(v2) / (self.length() * v2.length())

    def angle(self, other):
        """
        向量夹角
        :param other:
        :return:
        """
        return np.arccos(self.cos_angle(other))

    def is_vertical(self, other, eps=1e-8):
        """
        判断垂直
        :param other:
        :param eps:
        :return:
        """
        return equals_zero_all(self * other, eps=eps)

    def is_parallel(self, other, eps):
        """
        判断共线
        :return:
        """
        if isinstance(other, Vector):
            nv2 = other.x
        else:
            nv2 = np.asarray(other)
        nv1 = self.x
        return np.rank([nv1, nv2]) == 1

    def is_unit_vector(self, eps=1e-8):
        """
        是否是单位向量
        :return:
        """
        return -eps < self.length() - 1 < eps

    def norm(self):
        """
        返回其单位向量
        :return:
        """
        return Vector(self.x / self.length())

    def to_norm(self):
        """
        转化到单位向量
        :return:
        """
        self.x /= self.length()
        return self


class Vector2(Vector):
    @staticmethod
    def get_angle_pi(a, b):
        """
        a, b 为一维向量， 多点会错！！！
        返回 向量a和b之间的夹角， 值域是 -pi ~ pi
        """
        a = npa(a)
        b = npa(b)
        norm_a = npl.norm(a, axis=1) if len(a.shape) == 2 else npl.norm(a)
        norm_b = npl.norm(b, axis=1) if len(b.shape) == 2 else npl.norm(b)
        # 单位化（可以不用这一步）
        # a = a / norm_a  # 不能写成 a /= norm_a
        # b = b / norm_b  # 不能写成 b /= norm_b
        # 夹角cos值
        cos_ = np.dot(a, b) / (norm_a * norm_b)
        # 夹角sin值
        sin_ = np.cross(a, b) / (norm_a * norm_b)
        arctan2_ = np.arctan2(sin_, cos_)
        return arctan2_

    @staticmethod
    def get_angle(a: (list, np.ndarray), b: (list, np.ndarray)):
        """
        返回量向量之间的夹角，值域是 0~180，单位是度
        """
        # 初始化向量
        degree = np.rad2deg(Vector2.get_angle_pi(a, b))
        return np.abs(degree)

    @staticmethod
    def get_rotate_angle(v1, v2):
        """
        v1旋转到v2经历的角度， 值域0~360, 单位是度
        """
        return np.rad2deg(Vector2.get_angle_pi(v1, v2)) % 360



class Vector3(Vector):
    pass


class UnitVector(Vector):
    def __init__(self, x=None, dim=None, start=None, end=None, eps=1e-8):
        super(UnitVector, self).__init__(x, dim, start, end)
        assert self.is_unit_vector()


class UnitVector3(Vector3):
    def __init__(self, x=None, dim=None, start=None, end=None, eps=1e-8):
        super(UnitVector3, self).__init__(x, dim, start, end)
        assert self.is_unit_vector()


UNITVECTORX = UnitVector3((1, 0, 0))
UNITVECTORY = UnitVector3((0, 1, 0))
UNITVECTORZ = UnitVector3((0, 0, 1))


class Graphic(metaclass=ABCMeta):
    pass


class Point(Graphic):
    def __init__(self, p):
        if isinstance(p, Point):
            self.p = p
        else:
            self.p = npa(p)

    def __setitem__(self, instance, value):
        if isinstance(isinstance(instance, Number)):
            self.p[instance] = value
        else:
            raise ValueError('键值异常')

    def __getitem__(self, item):
        return self.p[item]

    def __eq__(self, other):
        if isinstance(other, np.ndarray):
            # TODO: check it
            if len(self.p) != len(other):
                return False
            else:
                return equals_zero_all(self.p - other)
        elif isinstance(other, Point):
            if len(self.p) != len(other.p):
                return False
            else:
                return equals_zero_all(self.p - other.p)

    def __sub__(self, other):
        if isinstance(other, Point):
            return Vector(self.p - other.p)
        else:
            raise ValueError('Error in Point')

    def __str__(self):
        itor = [str(_) for _ in self.p]
        return "Point <" + ', '.join(itor) + ">"


class LineString(Graphic):
    def __init__(self, X=None):
        if isinstance(X, LineString):
            self.X = X.X
        else:
            if len(X) == 0:
                X = []
            elif isinstance(X[0], Point):
                X = [each.p for each in X]
            self.X = npa(X)

    def length(self, metric='euclidean', *args, **kwargs):
        m, n = self.X.shape
        return np.sum(
            [pdist(self.X[ind:ind + 2, :], metric=metric, *args, **kwargs) for ind in range(m - 1)])

    def is_ring(self, eps=1e-8):
        """
        判断是否成环
        :return:
        """
        return equals_zero_all(self.X[0, :] - self.X[-1, :], eps=eps)


class Shape(Graphic):
    @abstractmethod
    def area(self):
        """
        面积
        :return:
        """
        pass

    @abstractmethod
    def girth(self):
        """
        周长
        :return:
        """
        pass


class Stereographic(Graphic):
    @abstractmethod
    def surface_area(self):
        """
        表面积
        :return:
        """
        pass

    @abstractmethod
    def volume(self):
        """
        体积
        :return:
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: graphics .py
# @time: 2020/9/6 23:48
# @Software: PyCharm


import unittest
from astartool.project import std_logging


class TestLine2D(unittest.TestCase):
    @classmethod
    @std_logging()
    def setup_class(cls):
        pass

    @classmethod
    @std_logging()
    def teardown_class(cls):
        print('teardown_class()')

    @std_logging()
    def setup_method(self, method):
        pass

    @std_logging()
    def teardown_method(self, method):
        pass






import configparser
import psycopg2


cp = configparser.ConfigParser()


class PGConnection:

    def __init__(self, session, file_name='db.conf'):
        cp.read(file_name)
        self.database = cp.get(session, 'database')
        self.user = cp.get(session, 'user')
        self.password = cp.get(session, 'password')
        self.host = cp.get(session, 'host')
        self.port = int(cp.get(session, 'port'))
        self.connection = None

    def __enter__(self):
        self.connection = psycopg2.connect(
            "dbname='%(dbname)s' user='%(user)s' password='%(password)s' port='%(port)s' host='%(host)s'" % {
                'dbname': self.database, 'user': self.user, 'password': self.password, 'port': self.port, 'host': self.host})

        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.connection is not None:
            self.connection.close()
